"""
Object inspector.

"""
import inspect
import re
import sys

from . import output, tools
from .exceptions import SeeError
from .features import FEATURES


class DefaultArg(object):
    """
    A global instance of this class is used as the default argument to
    :func:`see.see`. This allows for a distinction between calling ``see()``
    without arguments and calling it with a falsey argument like ``see(None)``.
    """
    def __repr__(self):
        return 'anything'


DEFAULT_ARG = DefaultArg()


class Namespace(object):
    """
    An object whose attributes are initialised with a dictionary.
    Similar to the SimpleNamespace_ class introduced in Python 3.

    .. _SimpleNamespace:  https://
       docs.python.org/3/library/types.html#types.SimpleNamespace
    """
    def __init__(self, attrs):
        self.__dict__.update(attrs)


# Get all the 'is' functions from the inspect module.
INSPECT_FUNCS = tuple((name, getattr(inspect, name))
                      for name in dir(inspect) if name.startswith('is'))


def handle_deprecated_args(tokens, args, kwargs):
    """
    Backwards compatibility with deprecated arguments ``pattern`` and ``r``.
    """
    num_args = len(args)
    pattern = args[0] if num_args else kwargs.get('pattern', None)
    regex = args[1] if num_args > 1 else kwargs.get('r', None)

    if pattern is not None:
        tokens = tools.filter_wildcard(tokens, pattern)
        sys.stderr.write(
            'Please use see().match() now. The "pattern" argument is '
            'deprecated and will be removed in a later release. \n')

    if regex is not None:
        tokens = tools.filter_regex(tokens, re.compile(regex))
        sys.stderr.write(
            'Please use see().regex() now. The "r" argument is '
            'deprecated and will be removed in a later release. \n')

    return tokens


def see(obj=DEFAULT_ARG, *args, **kwargs):
    """
    see(obj=anything)

    Show the features and attributes of an object.

    This function takes a single argument, ``obj``, which can be of any type.
    A summary of the object is printed immediately in the Python interpreter.
    For example::

        >>> see([])
            []            in            +             +=            *
            *=            <             <=            ==            !=
            >             >=            dir()         hash()
            help()        iter()        len()         repr()
            reversed()    str()         .append()     .clear()
            .copy()       .count()      .extend()     .index()
            .insert()     .pop()        .remove()     .reverse()
            .sort()

    If this function is run without arguments, it will instead list the objects
    that are available in the current scope. ::

        >>> see()
            os        random    see()     sys

    The return value is an instance of :class:`SeeResult`.
    """
    use_locals = obj is DEFAULT_ARG

    if use_locals:
        # Get the local scope from the caller's stack frame.
        # Typically this is the scope of an interactive Python session.
        obj = Namespace(inspect.currentframe().f_back.f_locals)

    tokens = []
    attrs = dir(obj)

    if not use_locals:

        for name, func in INSPECT_FUNCS:
            if func(obj):
                tokens.append(name)

        for feature in FEATURES:
            if feature.match(obj, attrs):
                tokens.append(feature.symbol)

    for attr in filter(lambda a: not a.startswith('_'), attrs):
        try:
            prop = getattr(obj, attr)
        except (AttributeError, Exception):  # pylint: disable=broad-except
            prop = SeeError()
        action = output.display_name(name=attr, obj=prop, local=use_locals)
        tokens.append(action)

    if args or kwargs:
        tokens = handle_deprecated_args(tokens, args, kwargs)

    return output.SeeResult(tokens)



"""
Unit tests for the see.term module.

"""
try:
    import unittest2 as unittest
except ImportError:
    import unittest

from see import term


class TestTermModule(unittest.TestCase):

    def test_line_width(self):
        # Arrange
        default_width = 1
        max_width = 1

        # Act
        width = term.line_width(default_width, max_width)
        width_no_args = term.line_width()

        # Assert
        self.assertIsInstance(width, int)
        self.assertEqual(width, 1)
        self.assertLessEqual(width_no_args, term.MAX_LINE_WIDTH)



"""
Unit tests for the see.output module.

"""
import re

try:
    import unittest2 as unittest
except ImportError:
    import unittest

from see import output, see
from see.features import PY3


class ObjectWithLongAttrName(dict):

    def lorem_ipsum_dolor_sit_amet_consectetur_adipiscing_elit(self):
        pass


class TestSeeResultClass(unittest.TestCase):

    def test_see_returns_result_instance(self):
        result_for_obj = see([])
        result_for_locals = see()

        isinstance(result_for_obj, output.SeeResult)
        isinstance(result_for_locals, output.SeeResult)

    def test_repr(self):
        result = see()

        self.assertEqual(str(result), result.__repr__())

    def test_acts_like_tuple(self):
        result = see([])

        self.assertIsNotNone(tuple(result))
        self.assertIsNotNone(result[0])
        self.assertIsNotNone(result[-1])
        self.assertTrue(any(result))
        self.assertTrue(len(result) > 0)

    def test_justify_attributes(self):
        obj = ObjectWithLongAttrName()

        result = see(obj)
        col_width = output.column_width(result)
        padded = [output.justify_token(tok, col_width) for tok in result]
        lens = sorted(map(len, padded))
        factors = tuple(float(num) / lens[0] for num in lens[1:])

        # Assert
        self.assertNotEqual(lens[0], lens[-1],
                            'Expected differing column widths')
        self.assertTrue(any(factors))
        self.assertTrue(all(int(factor) == factor for factor in factors),
                        'Irregular column widths')

    def test_filter_with_wildcard(self):
        obj = []
        pattern = '*op*'
        expected = ('.copy()', '.pop()') if PY3 else ('.pop()',)

        filtered_result = see(obj).filter(pattern)

        self.assertIsInstance(filtered_result, output.SeeResult)
        self.assertEqual(expected, filtered_result)

    def test_filter_with_regex_string(self):
        obj = []
        pattern = '/[aeiou]{2}/'
        expected = ('.clear()', '.count()') if PY3 else ('.count()',)

        filtered_result = see(obj).filter(pattern)

        self.assertIsInstance(filtered_result, output.SeeResult)
        self.assertEqual(expected, filtered_result)

    def test_filter_with_regex_object(self):
        obj = []
        pattern = re.compile('[aeiou]{2}')
        expected = ('.clear()', '.count()') if PY3 else ('.count()',)

        filtered_result = see(obj).filter(pattern)

        self.assertIsInstance(filtered_result, output.SeeResult)
        self.assertEqual(expected, filtered_result)



"""
Exception classes.

"""


class SeeError(Exception):
    """
    This is used internally by :func:`see.see` to indicate an invalid
    attribute, such as one that raises an exception when it is accessed.
    """
    pass



"""
see: dir for humans.

Documentation is available at https://araile.github.io/see/

"""
from .inspector import see
from .output import SeeResult


__all__ = ['see', 'SeeResult']

__author__ = 'Liam Cooke'
__contributors__ = 'See AUTHORS.rst'
__version__ = '1.4.0'
__copyright__ = 'Copyright (c) 2009-2017 Liam Cooke'
__license__ = 'BSD License'



"""
Manipulating strings for output.

"""
import math
import re
import sys
import textwrap

from . import term, tools
from .exceptions import SeeError
from .features import PY3


REGEX_TYPE = type(re.compile('.'))


class SeeResult(object):
    """
    The output of the :func:`see` function.

    Acts like a tuple of strings, so you can iterate over the output::

        >>> first = see()[0]

        >>> for string in see([]):
        ...     print(string)

    """
    def __init__(self, tokens):
        self._tokens = tuple(tokens)

    def __repr__(self):
        col_width = column_width(self)
        padded = [justify_token(tok, col_width) for tok in self]

        ps1 = getattr(sys, 'ps1', None)
        if ps1:
            get_len = tools.display_len if PY3 else len
            indent = ' ' * get_len(ps1)
        else:
            indent = '    '

        return textwrap.fill(''.join(padded), term.line_width(),
                             initial_indent=indent,
                             subsequent_indent=indent)

    def __eq__(self, other):
        return tuple(self) == tuple(other)

    def __iter__(self):
        return iter(self._tokens)

    def __len__(self):
        return len(self._tokens)

    def __getitem__(self, index):
        return self._tokens[index]

    def filter(self, pattern):
        """
        Filter the results using a pattern.

        This accepts a shell-style wildcard pattern (as used by the fnmatch_
        module)::

            >>> see([]).filter('*op*')
                .copy()    .pop()

        It also accepts a regular expression. This may be a compiled regular
        expression (from the re_ module) or a string that starts with a ``/``
        (forward slash) character::

            >>> see([]).filter('/[aeiou]{2}/')
                .clear()    .count()

        .. _fnmatch: https://docs.python.org/3/library/fnmatch.html
        .. _re: https://docs.python.org/3/library/re.html
        """
        if isinstance(pattern, REGEX_TYPE):
            func = tools.filter_regex
        elif pattern.startswith('/'):
            pattern = re.compile(pattern.strip('/'))
            func = tools.filter_regex
        else:
            func = tools.filter_wildcard

        return SeeResult(func(self, pattern))


def column_width(tokens):
    """
    Return a suitable column width to display one or more strings.
    """
    get_len = tools.display_len if PY3 else len
    lens = sorted(map(get_len, tokens or [])) or [0]
    width = lens[-1]

    # adjust for disproportionately long strings
    if width >= 18:
        most = lens[int(len(lens) * 0.9)]
        if most < width + 6:
            return most

    return width


def justify_token(tok, col_width):
    """
    Justify a string to fill one or more columns.
    """
    get_len = tools.display_len if PY3 else len
    tok_len = get_len(tok)
    diff_len = tok_len - len(tok) if PY3 else 0

    cols = (int(math.ceil(float(tok_len) / col_width))
            if col_width < tok_len + 4 else 1)

    if cols > 1:
        return tok.ljust((col_width * cols) + (4 * cols) - diff_len)
    else:
        return tok.ljust(col_width + 4 - diff_len)


def display_name(name, obj, local):
    """
    Get the display name of an object.

    Keyword arguments (all required):

    * ``name`` -- the name of the object as a string.
    * ``obj`` -- the object itself.
    * ``local`` -- a boolean value indicating whether the object is in local
      scope or owned by an object.

    """
    prefix = '' if local else '.'

    if isinstance(obj, SeeError):
        suffix = '?'
    elif hasattr(obj, '__call__'):
        suffix = '()'
    else:
        suffix = ''

    return ''.join((prefix, name, suffix))



"""
Unit tests for the see.inspector module.

"""
try:
    import unittest2 as unittest
except ImportError:
    import unittest

from see import inspector, output, see


class ObjectWithAttributeError(object):

    @property
    def bad_attribute(self):
        raise Exception


class ObjectWithDocstring(object):
    """
    Hello, world
    """
    pass


class ObjectWithEmptyDocstring(object):
    """
    """
    pass


class TestDefaultArgClass(unittest.TestCase):

    def test_global_instance(self):
        self.assertIsInstance(inspector.DEFAULT_ARG, inspector.DefaultArg)

    def test_repr(self):
        arg = inspector.DefaultArg()

        self.assertEqual(repr(arg), 'anything')


class TestNamespaceClass(unittest.TestCase):

    def test_namespace(self):
        attributes = {
            'hello': 'hello world',
        }

        namespace = inspector.Namespace(attributes)

        self.assertEqual(namespace.hello, 'hello world')
        self.assertIn('hello', dir(namespace))
        self.assertIn('.hello', see(namespace))


class TestSeeFunction(unittest.TestCase):

    def test_see_with_args(self):
        result = see([])

        self.assertTrue(any(result))

    def test_see_without_args(self):
        result = see()

        self.assertTrue(any(result))

    def test_see_with_new_default_arg_instance(self):
        result_with_obj = see(inspector.DefaultArg())
        result_without_args = see()

        self.assertNotEqual(result_with_obj, result_without_args)

    def test_see_accessor_raises_exception(self):
        normal_obj = 1
        err_obj = ObjectWithAttributeError()

        normal_result = see(normal_obj)
        err_result = see(err_obj)

        self.assertTrue(not any(attr.endswith('?')
                                for attr in normal_result))
        self.assertTrue(any(attr.endswith('?')
                            for attr in err_result))
        self.assertIn('.bad_attribute?', err_result)

    def test_see_object_has_help(self):
        obj_help = ObjectWithDocstring()
        obj_nohelp = ObjectWithEmptyDocstring()

        out_help = see(obj_help)
        out_nohelp = see(obj_nohelp)

        self.assertTrue('help()' in out_help)
        self.assertFalse('help()' in out_nohelp)

    # Deprecated filtering API -- see test_mod_output

    def test_see_deprecated_r_arg(self):
        obj = []
        pattern = '[aeiou]{2}'

        filtered_result = see(obj, r=pattern)
        filtered_result_pos_arg = see(obj, '*', pattern)

        # Assert
        self.assertIn('.count()', filtered_result)
        self.assertNotIn('.pop()', filtered_result)

    def test_see_deprecated_pattern_arg(self):
        obj = []
        pattern = '*()'

        filtered_result = see(obj, pattern=pattern)
        filtered_result_pos_arg = see(obj, pattern)

        # Assert
        self.assertEqual(filtered_result, filtered_result_pos_arg)
        self.assertIn('.count()', filtered_result)
        self.assertNotIn('+=', filtered_result)



"""
Terminal info.

"""
import platform
import struct

# pylint: disable=invalid-name

try:
    if platform.system() == 'Windows':  # pragma: no cover (windows)
        from ctypes import windll, create_string_buffer
        fcntl, termios = None, None
    else:
        import fcntl
        import termios
        windll, create_string_buffer = None, None
except ImportError:
    fcntl, termios = None, None
    windll, create_string_buffer = None, None


DEFAULT_LINE_WIDTH = 78
MAX_LINE_WIDTH = 120


def term_width():
    """
    Return the column width of the terminal, or ``None`` if it can't be
    determined.
    """
    if fcntl and termios:
        try:
            winsize = fcntl.ioctl(0, termios.TIOCGWINSZ, '    ')
            _, width = struct.unpack('hh', winsize)
            return width
        except IOError:
            pass
    elif windll and create_string_buffer:  # pragma: no cover (windows)
        stderr_handle, struct_size = -12, 22
        handle = windll.kernel32.GetStdHandle(stderr_handle)
        csbi = create_string_buffer(struct_size)
        res = windll.kernel32.GetConsoleScreenBufferInfo(handle, csbi)
        if res:
            (_, _, _, _, _, left, _, right, _,
             _, _) = struct.unpack('hhhhHhhhhhh', csbi.raw)
            return right - left + 1


def line_width(default_width=DEFAULT_LINE_WIDTH, max_width=MAX_LINE_WIDTH):
    """
    Return the ideal column width for the output from :func:`see.see`, taking
    the terminal width into account to avoid wrapping.
    """
    width = term_width()
    if width:  # pragma: no cover (no terminal info in Travis CI)
        return min(width, max_width)
    else:
        return default_width



"""
Python feature definitions.

Reference links:

  * `What's New in Python
    <https://docs.python.org/3/whatsnew/index.html>`__

  * `Data model
    <https://docs.python.org/3/reference/datamodel.html>`__

  * `Standard operators as functions
    <https://docs.python.org/3/library/operator.html>`__

"""
# -----------------------------------------------------------------------------
# Relevant Python changes
#
# 3.5
# - Introduced the @ operator and __matmul__, __imatmul__
# - Introduced __await__, __aiter__, __anext__, __aenter__, __aexit__
#   (see PEP 492)
#
# 3.0-3.3
# - Dropped __cmp__
# - Dropped __div__
# -----------------------------------------------------------------------------

import sys

from .tools import compact


PY_VERSION = sys.version_info

if (PY_VERSION < (2, 6)) or ((3, 0) <= PY_VERSION < (3, 3)):  # pragma: nocover
    sys.stderr.write('Warning: see() is not supported here. '
                     'Please use Python 3.3+ or 2.6+.\n')

PY2 = PY_VERSION < (3, 0)
PY3 = not PY2
PY3_5 = (3, 5) <= PY_VERSION


class Feature(object):
    """
    Definition of a Python feature that an object might support (such as
    performing an arithmetic operation or returning a length), and a symbol
    to indicate this in the output from :func:`see.see`.

    >>> add_feature = Feature(symbol='+', attrs=[
    ...     '__add__',
    ...     '__radd__',
    ... ])

    Support for this feature is detected by checking for one or more special
    attributes on the object.
    """

    def __init__(self, symbol, attrs=None):
        self.symbol = symbol
        self.attrs = set(compact(attrs))

    def match(self, obj, attrs):
        """
        Return whether the feature represented by this symbol is supported by
        a given object and its attributes.
        """
        # pylint: disable=unused-argument
        return not self.attrs.isdisjoint(attrs)


class HelpFeature(Feature):
    """
    Help feature.
    """

    def match(self, obj, attrs):
        """
        Only match if the object contains a non-empty docstring.
        """
        if '__doc__' in attrs:
            lstrip = getattr(obj.__doc__, 'lstrip', False)
            return lstrip and any(lstrip())


FEATURES = compact((

    # -------------------------------------------------------------------------
    # Callable

    Feature(symbol='()', attrs=['__call__']),

    # -------------------------------------------------------------------------
    # Element/attribute access

    Feature(symbol='.*', attrs=['__getattr__']),
    Feature(symbol='[]', attrs=['__getitem__', '__setitem__', '__delitem__']),

    # -------------------------------------------------------------------------
    # Iteration

    Feature(symbol='with', attrs=['__enter__', '__exit__']),
    Feature(symbol='in', attrs=['__contains__']),

    # -------------------------------------------------------------------------
    # Operators

    Feature(symbol='+', attrs=['__add__', '__radd__']),
    Feature(symbol='+=', attrs=['__iadd__']),

    Feature(symbol='-', attrs=['__sub__', '__rsub__']),
    Feature(symbol='-=', attrs=['__isub__']),

    Feature(symbol='*', attrs=['__mul__', '__rmul__']),
    Feature(symbol='*=', attrs=['__imul__']),

    Feature(symbol='@', attrs=['__matmul__', '__rmatmul__']) if PY3_5 else 0,
    Feature(symbol='@=', attrs=['__imatmul__']) if PY3_5 else 0,

    # Python 2.x uses __div__ for a/b, but will look for __truediv__ instead
    # if this is enabled with `from __future__ import division`.
    Feature(symbol='/', attrs=[
        '__truediv__',
        '__rtruediv__',
        PY2 and '__div__',
        PY2 and '__rdiv__',
    ]),
    Feature(symbol='/=', attrs=[
        '__itruediv__',
        PY2 and '__idiv__',
    ]),

    Feature(symbol='//', attrs=['__floordiv__', '__rfloordiv__']),
    Feature(symbol='//=', attrs=['__ifloordiv__']),

    Feature(symbol='%', attrs=['__mod__', '__rmod__', '__divmod__']),
    Feature(symbol='%=', attrs=['__imod__']),

    Feature(symbol='**', attrs=['__pow__', '__rpow__']),
    Feature(symbol='**=', attrs=['__ipow__']),

    Feature(symbol='<<', attrs=['__lshift__', '__rlshift__']),
    Feature(symbol='<<=', attrs=['__ilshift__']),

    Feature(symbol='>>', attrs=['__rshift__', '__rrshift__']),
    Feature(symbol='>>=', attrs=['__irshift__']),

    Feature(symbol='&', attrs=['__and__', '__rand__']),
    Feature(symbol='&=', attrs=['__iand__']),

    Feature(symbol='^', attrs=['__xor__', '__rxor__']),
    Feature(symbol='^=', attrs=['__ixor__']),

    Feature(symbol='|', attrs=['__or__', '__ror__']),
    Feature(symbol='|=', attrs=['__ior__']),

    Feature(symbol='+obj', attrs=['__pos__']),
    Feature(symbol='-obj', attrs=['__neg__']),
    Feature(symbol='~', attrs=['__invert__']),

    Feature(symbol='<', attrs=['__lt__', PY2 and '__cmp__']),
    Feature(symbol='<=', attrs=['__le__', PY2 and '__cmp__']),
    Feature(symbol='==', attrs=['__eq__', PY2 and '__cmp__']),
    Feature(symbol='!=', attrs=['__ne__', PY2 and '__cmp__']),
    Feature(symbol='>', attrs=['__gt__', PY2 and '__cmp__']),
    Feature(symbol='>=', attrs=['__ge__', PY2 and '__cmp__']),

    # -------------------------------------------------------------------------
    # Builtin functions

    Feature(symbol='abs()', attrs=['__abs__']),
    Feature(symbol='bool()', attrs=[
        '__bool__' if PY3 else '__nonzero__'
    ]),
    Feature(symbol='complex()', attrs=['__complex__']),
    Feature(symbol='dir()', attrs=['__dir__']) if PY3 else 0,
    Feature(symbol='divmod()', attrs=['__divmod__', '__rdivmod__']),
    Feature(symbol='float()', attrs=['__float__']),
    Feature(symbol='hash()', attrs=['__hash__']),
    HelpFeature(symbol='help()'),
    Feature(symbol='hex()', attrs=[
        '__index__' if PY3 else '__hex__'
    ]),
    Feature(symbol='int()', attrs=['__int__']),
    Feature(symbol='iter()', attrs=['__iter__']),
    Feature(symbol='len()', attrs=['__len__']),
    Feature(symbol='long()', attrs=['__long__']) if PY2 else 0,
    Feature(symbol='oct()', attrs=[
        '__index__' if PY3 else '__oct__'
    ]),
    Feature(symbol='repr()', attrs=['__repr__']),
    Feature(symbol='reversed()', attrs=['__reversed__']),
    Feature(symbol='round()', attrs=['__round__']) if PY3 else 0,
    Feature(symbol='str()', attrs=['__str__']),
    Feature(symbol='unicode()', attrs=['__unicode__']) if PY3 else 0,

))



"""
Unit tests for the see.tools module.

"""
import re

try:
    import unittest2 as unittest
except ImportError:
    import unittest

from see import tools


class TestToolsModule(unittest.TestCase):

    def test_compact(self):
        # Arrange
        items = [0, 1, '', 'foo', True, False, [], ['a'], {}, {1: 2}]
        expected = (1, 'foo', True, ['a'], {1: 2})

        # Act
        compacted = tools.compact(items)

        # Assert
        self.assertEqual(compacted, expected)

    def test_filter_regex(self):
        # Arrange
        names = ["george", "helen"]
        pat_wildcard = re.compile("e.*g")
        pat_start = re.compile("^h")
        pat_end = re.compile("n$")

        # Act
        out_wildcard = tools.filter_regex(names, pat_wildcard)
        out_start = tools.filter_regex(names, pat_start)
        out_end = tools.filter_regex(names, pat_end)

        # Assert
        self.assertIsInstance(out_wildcard, tuple)
        self.assertIsInstance(out_start, tuple)
        self.assertIsInstance(out_end, tuple)
        self.assertEqual(out_wildcard, ("george",))
        self.assertEqual(out_start, ("helen",))
        self.assertEqual(out_end, ("helen",))

    def test_filter_wildcard(self):
        # Arrange
        names = ["george", "helen"]
        pat_wildcard = "*or*"
        pat_single = "h?l?n"
        pat_partial = "e*"

        # Act
        out_wildcard = tools.filter_wildcard(names, pat_wildcard)
        out_single = tools.filter_wildcard(names, pat_single)
        out_partial = tools.filter_wildcard(names, pat_partial)

        # Assert
        self.assertIsInstance(out_wildcard, tuple)
        self.assertIsInstance(out_single, tuple)
        self.assertIsInstance(out_partial, tuple)
        self.assertEqual(out_wildcard, ("george",))
        self.assertEqual(out_single, ("helen",))
        self.assertEqual(out_partial, ())



"""
Filtering and other tasks.

"""
import fnmatch
import unicodedata


def compact(objects):
    """
    Filter out any falsey objects in a sequence.
    """
    return tuple(filter(bool, objects or []))


def char_width(char):
    """
    Get the display length of a unicode character.
    """
    if ord(char) < 128:
        return 1
    elif unicodedata.east_asian_width(char) in ('F', 'W'):
        return 2
    elif unicodedata.category(char) in ('Mn',):
        return 0
    else:
        return 1


def display_len(text):
    """
    Get the display length of a string. This can differ from the character
    length if the string contains wide characters.
    """
    text = unicodedata.normalize('NFD', text)
    return sum(char_width(char) for char in text)


def filter_regex(names, regex):
    """
    Return a tuple of strings that match the regular expression pattern.
    """
    return tuple(name for name in names
                 if regex.search(name) is not None)


def filter_wildcard(names, pattern):
    """
    Return a tuple of strings that match a shell-style wildcard pattern.
    """
    return tuple(name for name in names
                 if fnmatch.fnmatch(name, pattern))
